# Position Explorer
# Copyright 2010 Glenn Babcock

# Update 12 Feb 2010
# Change log
# - added ability to get the global positon of vertices in groups/components
# - improved formatting of output
# - added vertices only option

# Update 11 Feb 2010 by Glenn Babcock
# Change log
# - improved formatting of output
# - added "no faces" option

# Initial Release 10 Feb 2010 by Glenn Babcock
# - analyze the position and other information about selected entities
# - supports groups, nested groups, faces, edges, vertices
# - report on single or multiple selections
# - recursive feature allows reporting of group entities
# - outputs to Ruby Console

# program flow
# - define classes
# - if not already, load menus and create instance of PositionReporter
# - on menu, Do PositionReporter.getPositionReport with selection and options
# - PositionReporter.getPositionReport creates an instance of AnalyzeEntities and calls getEntityReport with entities from selection
# - AnalyzeEntities.getEntityReport iterates though selection entities, calling other Type-oriented classes as needed
# - Each Reporter builds a string containing the report, which is the output to the terminal

require 'sketchup.rb'

version_required = 7.0
if (Sketchup.version.to_f < version_required) 
  UI.messagebox("You must have Sketchup " + version_required.to_s +
                " to run this plugin. Visit sketchup.google.com to upgrade.")
  return
end

module DebugPE
	attr :debugPE
	attr :debugPE1
	def getFlags
		@debugPE = false
		@debugPE1 = false
		@debugPE2 = false
	end
end
#=====================================================================================================
#=====================================================================================================
class PositionReporter
	include DebugPE
	def initialize
		getFlags
	end
	
	def getPositionReport(sel, recursive, faces, detail, vertices)
		level=0
		str=""
		str+="\n***** Begin Position Report #{Time.now.inspect} *****\n"
		str+="Options: Recursive(#{recursive}), Faces(#{faces}), Detail(#{detail})\n"
		sel.each{|ent|
			if vertices
				ae=AnalyzeEntities.new
				str+=ae.getVerticesReport(ent, recursive, level)
			else
				sel.make_unique if sel.is_a? Sketchup::Group #addresses SU bug with group copies
				ae=AnalyzeEntities.new
				str+=ae.getEntityReport(ent, recursive, faces, detail, level)
			end
		}
		str+="***** End Position Report #{Time.now.inspect} *****\n"
		puts str
	end
end

class AnalyzeEntities
	include DebugPE
	def initialize
		getFlags
	end
	
	def getEntityReport(ent, recursive, faces, detail, level)
		str=""
		if ent.is_a? Sketchup::Group || Sketchup::ComponentInstance
			ent.make_unique if ent.is_a? Sketchup::Group #addresses SU bug with group copies
			groupReport=AnalyzeGroup.new
			str+=groupReport.getGroupReport(ent, recursive, faces, detail, level)
		elsif ent.is_a? Sketchup::Edge
			edgeReport=AnalyzeEdge.new
			str+=edgeReport.getEdgeReport(ent, recursive, detail, level)			
		elsif ent.is_a? Sketchup::Face
			if faces
				edgeReport=AnalyzeEdge.new
				str+=edgeReport.getEdgeReport(ent, recursive, detail, level)	
			end
		else
			str+="Type #{ent.typename} not supported\n"
		end
		return str
	end
	
	def getVerticesReport(ent, recursive, level)
		str=""
		if ent.is_a? Sketchup::Group
			ent.make_unique if ent.is_a? Sketchup::Group #addresses SU bug with group copies
			groupReport=AnalyzeGroup.new
			str+=groupReport.getGroupVerticesReport(ent, recursive, level)
		elsif ent.is_a? Sketchup::ComponentInstance
			componentReport=AnalyzeComponent.new
			str+=componentReport.getComponentVerticesReport(ent, recursive, level)
		elsif ent.is_a? Sketchup::Edge
			edgeReport=AnalyzeEdge.new
			str+=edgeReport.getEdgeVerticesReport(ent, recursive, level)			
		elsif ent.is_a? Sketchup::Face
			edgeReport=AnalyzeEdge.new
			str+=edgeReport.getEdgeVerticesReport(ent, recursive, level)	
		else
			str+="Type #{ent.typename} not supported\n"
		end
		return str
	end
end

class AnalyzeEdge
	include DebugPE
	def initialize
		getFlags
	end
	
	def getEdgeReport(ent, recursive, detail, level)
		v_global=VerticesGlobalPositions.new
		g_pos=v_global.getGlobalPosition(ent)
		
		tabs=Tabs.new
		tabstr=tabs.get(level)
		
		str=tabstr
		str+="#{ent.to_s}; ID:#{ent.entityID}\n"
			str+=tabstr if detail
		str+="Edge ID:#{ent.entityID}\n" if detail
			str+=tabstr if detail
		str+="Parent=#{ent.parent}\n" if detail
			str+=tabstr if detail
		str+="Length=#{ent.length}\n" if detail
		ent.vertices.each_with_index{|vert,i|
			str+=tabstr if detail
			str+="\t#{vert.to_s}, \n" if detail
				str+=tabstr
			str+="\tVert ID:#{vert.entityID} [#{i}].local pos  = #{vert.position}\n"
				str+=tabstr
			str+="\tVert ID:#{vert.entityID} [#{i}].global pos = #{g_pos[i]}\n"
		}
		return str
	end
	
	def getEdgeVerticesReport(ent, recursive, level)
		v_global=VerticesGlobalPositions.new
		g_pos=v_global.getGlobalPosition(ent)
		
		tabstr="\t"
		
		str=""
		ent.vertices.each_with_index{|vert,i|
				str+=tabstr
			str+="Vert ID:#{vert.entityID} [#{i}].local pos  = #{vert.position}\n"
				str+=tabstr
			str+="Vert ID:#{vert.entityID} [#{i}].global pos = #{g_pos[i]}\n"
		}
		return str
	end
end

class AnalyzeGroup
	include DebugPE
	def initialize
		getFlags
	end
	
	def getGroupReport(ent, recursive, faces, detail, level)
		tabs=Tabs.new
		tabstr=tabs.get(level)
		
		str=tabstr
		str+="\n"
			str+=tabstr
		str+="#{ent.to_s}; ID:#{ent.entityID}; Name=#{ent.name}; contains #{ent.entities.length} entities\n"
			str+=tabstr if detail
		str+="Parent=#{ent.parent}\n" if detail
		#get transformation
			str+=tabstr
		str+="Group Transformation:\n"
		ent_t=ent.transformation.to_a
		0.upto(3){|x|
			str+=tabstr
			0.upto(3){|y|
				str+="#{"%.3f" % ent_t[((x*4)+y)]}"
				str+="\t|\t" if y<3
			}
			str+="\n"
		}
		if recursive
			
				puts "AnalyzeGroup recursive level = #{level}" if @debugPE2
				
				str+=tabstr
			str+="Entities in Group ID:#{ent.entityID}:\n"

			level+=1
			tabstr=tabs.get(level)

			e_sort=ent.entities.sort { |a,b| a.to_s <=> b.to_s }
			e_sort.each{|sub_ent|
				ae=AnalyzeEntities.new
				str+=ae.getEntityReport(sub_ent, recursive, faces, detail, level)
			}
				str+=tabstr
			str+="End report on Entities in Group ID:#{ent.entityID}:\n\n"
		end
		return str
	end
	
	def getGroupVerticesReport(ent, recursive, level)
		tabstr="\t"
		
		str=tabstr
		if recursive
			level+=1

			e_sort=ent.entities.sort { |a,b| a.to_s <=> b.to_s }
			e_sort.each{|sub_ent|
				ae=AnalyzeEntities.new
				str+=ae.getVerticesReport(sub_ent, recursive, level)
			}
				str+=""
		end
	end
end

class Tabs
	include DebugPE
	def initialize
		getFlags
	end

	def get(level)
			puts "level = #{level}" if @debugPE1
		str=""
		1.upto(level){|x|
			str+="\t"
		}
		return str
	end
end

class VerticesGlobalPositions
	include DebugPE
	def initialize
		getFlags
	end
	
	#returns an array of Point3d objects that equal the global position of the vertices in an edge	
	#currently only gets the first parent instance, components may have many instances
	def getGlobalPosition(ent) 
		g_pos=[] #array to store global positions of vertices
		#get local position of vertices
		ent.vertices.each_with_index{|vert,i|
			g_pos[i]=vert.position
		}

		#get the parent of the edge
		p=ent.parent

		#while the parent is a group or component
		while p.is_a? Sketchup::ComponentDefinition
			#get the group/component transformation
			grp=p.instances[0]
			grp_t=grp.transformation

			#apply the group transformation to the stored position of the vertices
			g_pos.each{|g|
				g=g.transform! grp_t
			}
			
			#get the next parent in the hierarchy
			p=p.instances[0].parent
		end	
		return g_pos
	end
end

#=====================================================================================================
#=====================================================================================================
#initialize menus
if(!$position_Reporter_loaded)
	pos_reporter = PositionReporter.new
	
	#Right click UI menu.
	UI.add_context_menu_handler { |menu|
			
		menu=menu.add_submenu("Position Explorer")
			#pos_reporter.getPositionReport(Sketchup.active_model.selection,recursive,faces,detail)
			menu.add_item("Position Report"){pos_reporter.getPositionReport(Sketchup.active_model.selection,false,true,false,false)}
			menu.add_item("Position Report Recursive"){pos_reporter.getPositionReport(Sketchup.active_model.selection,true,true,false,false)}			
			menu.add_item("Position Report Recursive, No Faces"){pos_reporter.getPositionReport(Sketchup.active_model.selection,true,false,false,false)}
			menu.add_item("Position Report Recursive, Vertices Only"){pos_reporter.getPositionReport(Sketchup.active_model.selection,true,false,false,true)}
			menu.add_item("Position Report Detailed Recursive"){pos_reporter.getPositionReport(Sketchup.active_model.selection,true,true,true,false)}	
	}
	
	$position_Reporter_loaded=true
end


